
// --------------------------------------------------------------------------------
// ------- Add to "Post BUild Events in Case of moving to different folder" -------
// ECHO F | XCOPY /Y "....\DaySim.mexw64" "....\DaySim.mexw64"

// -----------------------------
// ------- CUDA Includes -------
#include "cuda_runtime.h"
#include "device_launch_parameters.h"
#include "cuda.h"

// ----------------------------
// ------- C++ Includes -------
#include "stdio.h"
#include "stdint.h"
#include "string.h"
#include "string"

// -------------------------------
// ------- Matlab Includes -------
#include <matrix.h>
#include <mex.h>
using namespace std;

#define DEBUG 0
#define HANDLE_CUDA( err )  (Handle_Error( err, __FILE__, __LINE__ ))

/******************************************************
 *@brief Error Handling Function of all the cuda Calls
 ****************************************************/
void Handle_Error(cudaError_t err, const char *file, int line) {
	if (err != cudaSuccess) {
		std::string Message = std::string(cudaGetErrorString(err)) + std::string(" in ") + std::string(file) + std::string(" at line: ") + std::to_string(line) + std::string("\n");
		mexErrMsgTxt(Message.c_str());
	}
}

/************************
 * Vector with 3 elements
 ***********************/
template <typename Type>
struct vec3 {
	Type X;
	Type Y;
	Type Z;
};



/********************************************
 *@brief Rotates point with a rotation Matrix
 *******************************************/
template <typename Type>
__device__ void Rotate(Type *Matrix, vec3<Type> *Vec) {
	Type V0 = Vec->X;
	Type V1 = Vec->Y;
	Type V2 = Vec->Z;

	Vec->X = V0 * Matrix[0] + V1 * Matrix[1] + V2 * Matrix[2];
	Vec->Y = V0 * Matrix[3] + V1 * Matrix[4] + V2 * Matrix[5];
	Vec->Z = V0 * Matrix[6] + V1 * Matrix[7] + V2 * Matrix[8];
}

/***********************
 *@brief Vector Addition
 ***********************/
template <typename Type>
__device__ const vec3<Type> operator+(const vec3<Type>& lhs, const vec3<Type>& rhs) {
	vec3<Type> Res;
	Res.X = lhs.X + rhs.X;
	Res.Y = lhs.Y + rhs.Y;
	Res.Z = lhs.Z + rhs.Z;
	return Res;
}

/**************************
 *@brief Vector Subtraction
 **************************/
template <typename Type>
__device__ const vec3<Type> operator-(const vec3<Type>& lhs, const vec3<Type>& rhs) {
	vec3<Type> Res;
	Res.X = lhs.X - rhs.X;
	Res.Y = lhs.Y - rhs.Y;
	Res.Z = lhs.Z - rhs.Z;
	return Res;

}

/*********************************
 *@brief Piecewise vector multiply
 ********************************/
template <typename Type>
__device__ const vec3<Type> operator*(const vec3<Type>& lhs, const vec3<Type>& rhs){
	vec3<Type> Res;
	Res.X = lhs.X * rhs.X;
	Res.Y = lhs.Y * rhs.Y;
	Res.Z = lhs.Z * rhs.Z;
	return Res;
}

/*********************************
 *@brief Piecewise vector multiply
 ********************************/
template <typename Type>
__device__ const vec3<Type> operator*(const Type& lhs, const vec3<Type>& rhs) {
	vec3<Type> Res;
	Res.X = lhs * rhs.X;
	Res.Y = lhs * rhs.Y;
	Res.Z = lhs * rhs.Z;
	return Res;
}

/******************
 *@brief Vector Sum
 *****************/
template <typename Type>
__device__ Type Vsum(vec3<Type> Data) {
	return (Data.X + Data.Y  + Data.Z);
}

/***********************************
 *@brief Converts Degress to Radians
 **********************************/
template <typename Type>
__device__ Type deg2rad(Type Deg) {
	Type Rad = Deg * ((Type)3.14159265358979) / ((Type)180.0);
	return Rad;

}



/*********************************************************
 *@brief Checks if the Sun is Visible from custom Location
 ********************************************************/
template <typename Type>
__device__ uint8_t SunVisible(vec3<Type> Point, vec3<Type> Sun,vec3<Type> Earth,Type Radius) {

	Type a = Vsum<Type>((Sun - Point)*(Sun - Point));
	Type b = ((Type)2.0) * Vsum<Type>((Sun - Point)*(Point - Earth));
	Type c = Vsum<Type>(Earth * Earth) + Vsum<Type>(Point * Point) - ((Type)2.0) * Vsum<Type>(Earth*Point) - Radius * Radius;
	Type Res = b * b - ((Type)4.0) * a*c;

	if (Res > 0.0) {


		Type R0 = (-b + sqrt(Res)) / (((Type)2.0) * a);
		Type R1 = (-b - sqrt(Res)) / (((Type)2.0) * a);

		vec3<Type> IP0 = Point + R0 * (Sun - Point);
		vec3<Type> IP1 = Point + R1 * (Sun - Point);

		// Have to Check whether the receiver is further from the sun - Night
		// Or whether the Intersection Point is closer to the Sun - Day
		Type D0 = sqrt(Vsum<Type>((IP0 - Sun)*(IP0 - Sun)));
		Type D1 = sqrt(Vsum<Type>((IP1 - Sun)*(IP1 - Sun)));
		Type DP = sqrt(Vsum<Type>((Point - Sun)*(Point - Sun)));

		Type Diff0 = fabs(D0 - DP);
		Type Diff1 = fabs(D1 - DP);

		if (Diff0 < Diff1) {

			if (D0 < D1) {
				// Receiver is at P0 and is Closer to the Sun
				// The Sun should be visible
				return 1;
			}
			else {
				// The Receiver is at P0, but this is further from the Sun
				// The Sun is not Visible
				return 0;
			}
		}
		else {
			if (D1 < D0) {
				// The Receiver is at P1 and is Closer to the Sun
				// The Sun is visible
				return 1;
			}
			else {
				//The Receiver is at P1, but is further from the Sun
				// The Sun is not Visible
				return 0;
			}
		}
	}else {
		// Likely Sunset or Sunrise
		return 1;
	}

}

/**************************************************
 *@brief computes rotation Matrix around the X Axis
 * NOTE: https://en.wikipedia.org/wiki/Rotation_matrix
 * 
 * [ 1   0        0     ]
 * [ 0 +cos(x)  +sin(x) ]
 * [ 0 -sin(x)  +cos(x) ]
 * 
 *************************************************/
template <typename Type>
__device__ void ComputeMatrix_X(Type *Matrix, Type Angle) {
	Matrix[0] = (Type)1.0;
	Matrix[1] = (Type)0.0;
	Matrix[2] = (Type)0.0;
	Matrix[3] = (Type)0.0;
	Matrix[4] = +cos(Angle);
	Matrix[5] = +sin(Angle);
	Matrix[6] = (Type)0.0;
	Matrix[7] = -sin(Angle);
	Matrix[8] = +cos(Angle);
}



/**************************************************
 *@brief computes rotation Matrix around the Z Axis 
 * NOTE: https://en.wikipedia.org/wiki/Rotation_matrix
 * 
 * [ +cos(x) -sin(x) 0]
 * [ +sin(x) +cos(x) 0]
 * [   0       0     1]
 * 
 *************************************************/
template <typename Type>
__device__ void ComputeMatrix_Z(Type *Matrix, Type Angle) {
	Matrix[0] = +cos(Angle);
	Matrix[1] = +sin(Angle);
	Matrix[2] = (Type)0.0;
	Matrix[3] = -sin(Angle);
	Matrix[4] = +cos(Angle);
	Matrix[5] = (Type)0.0;
	Matrix[6] = (Type)0.0;
	Matrix[7] = (Type)0.0;
	Matrix[8] = (Type)1.0;
}

/**********************************************************************
 *@brief Computing Kernel for Sun Visibility
 *@param EarthPos    : Position of the Earth in Time
 *@param Sun         : Position of the Sun - Expected to be [0,0,0]
 *@param Rec         : Custom position of the Receiver / Observer on the Earth
 *@param EarthRot    : Angle Definiting the Spin of the Earth in Time
 *@param EarthTilt   : Earth Axis Tilt (23.48 Deg)
 *@param EarthRadius : Radius of the Earth in Meters [6378 000 Merters]
 *@param N           : Amount of Samples
 *********************************************************************/
template <typename Type>
__global__ void Compute(vec3<Type> *EarthPos,vec3<Type> Sun,vec3<Type> Rec, Type *EarthRot, Type EarthTilt, Type EarthRadius,int N,uint8_t *Visible){
	unsigned int tix = threadIdx.x + blockIdx.x * blockDim.x;
	unsigned int tiy = threadIdx.y + blockIdx.y * blockDim.y;
	unsigned int tid = tix + tiy * blockIdx.x * gridDim.x;

	if (tid < (unsigned int)N) {
		vec3<Type> Position = Rec;                                // Fetch Original Position of the Observer
		vec3<Type> Ep = EarthPos[tid];                            // Fetch Earth Position in Time
		Type ZAngle = deg2rad<Type>(EarthRot[tid]);              // Convert From Degrees to Radians
		Type XAngle = deg2rad<Type>(EarthTilt);                  // Convert From Degress To Radians
		Type ZMatrix[9];
		Type XMatrix[9];
		ComputeMatrix_Z<Type>(ZMatrix, ZAngle);  // COmpute Host_Rotation Matrix around Z Axis
		ComputeMatrix_X<Type>(XMatrix, XAngle);  // Compute Host_Rotation Matrix around X Axis
		Rotate<Type>(ZMatrix, &Position);        // Earth Tilt Host_Rotation
		Rotate<Type>(XMatrix, &Position);        // Earth Host_Rotation in time

		// Superposition the Observer and Earth Position
		Position = Position + Ep;

		// Compute Visibility
		Visible[tid] = SunVisible<Type>(Position, Sun, Ep, EarthRadius); 
	}

 
}


/***************************************************************
 *@brief Transforms a Colum-Major Order Array to Row-Major Order
 **************************************************************/
template <typename Type>
void ColToRow(Type *input, Type *output, int Nx, int Ny) {
	int counter = 0;
	for (int i = 0; i < Ny; i++) {
		for (int j = 0; j < Nx; j++) {
			int index = j * Ny + i;
			output[counter] = input[index];
			counter++;
		}
	}
}

/****************************************************************
 *@brief Transforms a Row-Major Order Array to Column-Major Order
 ****************************************************************/
template <typename Type>
void RowToCol(Type *output, Type* input, int Nx, int Ny) {
	int counter = 0;
	for (int i = 0; i < Ny; i++) {
		for (int j = 0; j < Nx; j++) {
			int index = j * Ny + i;
			output[index] = input[counter];
			counter++;
		}
	}
}





/**************************************************
 *@brief Main Matlab Function
 *@param nlhs - Amount of LeftHandSide Parameters
 *@param nrhs - Amount of RightHandside Parameters
 *@param plhs - Array of Output Arguments
 *@param prhs - Array of Input Arguments
 ***************************************************/
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {

	const mxArray *M_EarthPos = prhs[0]; // Earth Positions in Time						[Length Based on Discretization in Time]
	const mxArray *M_Rotation = prhs[1]; // Rotation of the Earth in Time in Degress	[Length Based on Discretization in Time]
	const mxArray *M_SunPos   = prhs[2]; // Sun Center					[X,Y,Z]
	const mxArray* M_Observer = prhs[3]; // Simulated Point on Erath	[X,Y,Z]
	const mxArray *M_Args     = prhs[4]; // Additional Parameters [Earth Tilt | Earth Radius] (To Play Around with)


	/*********************************************************************************************
	 * NOTE: If the Library is compiled for architecture not present in the system, it will crash.
	 * https://arnon.dk/matching-sm-architectures-arch-and-gencode-for-various-nvidia-cards/
	 * NOTE: Debug build of this library is not used. Do not compile.
	 **********************************************************************************************/

	// ---------------------------------
	// -------- Check Parameters -------
	if (nrhs < 5) {
		std::string Message = std::string("Invalid amount of Arguments!\n");
		mexErrMsgTxt(Message.c_str());
	}

	// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
	if ((mxGetClassID(M_EarthPos) == mxDOUBLE_CLASS) && (mxGetClassID(M_Rotation) == mxDOUBLE_CLASS) && (mxGetClassID(M_SunPos) == mxDOUBLE_CLASS) && (mxGetClassID(M_Observer) == mxDOUBLE_CLASS)) {
	// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
		
		// ------------------------------------------
		// ------- Get Pointers to Data Arrays ------
		// NOTE: It is recommended to switch to mxGetDoubles() in newer generation Matlabs
		double *In_EarthPos = static_cast<double*>(mxGetData(M_EarthPos));
		double *In_Rot		= static_cast<double*>(mxGetData(M_Rotation));
		double *In_Sun		= static_cast<double*>(mxGetData(M_SunPos));
		double *In_Obs		= static_cast<double*>(mxGetData(M_Observer));
		double *In_Add		= static_cast<double*>(mxGetData(M_Args));

		vec3<double> Sun;
		vec3<double> Observer;
		vec3<double>  *Host_EarthPos = nullptr;
		double *Host_Rotation = nullptr;
		double EarthTilt   = 23.48;  // Degrees 
		double EarthRadius = 6378e3; // Meters

		Sun.X       = In_Sun[0];
		Sun.Y       = In_Sun[1];
		Sun.Z       = In_Sun[2];
		Observer.X  = In_Obs[0];
		Observer.Y  = In_Obs[1];
		Observer.Z  = In_Obs[2];
		EarthTilt   = In_Add[0];
		EarthRadius = In_Add[1];

		const mwSize* Dims = mxGetDimensions(M_EarthPos);
		int Nx = static_cast<int>(Dims[1]);
		int Ny = static_cast<int>(Dims[0]);

#if DEBUG
		
		mexPrintf("Nx: %d\n", Nx);
		mexPrintf("Ny: %d\n", Ny);
		mexPrintf("SunPos: [%f %f %f]\n", Sun.X, Sun.Y, Sun.Z);
		mexPrintf("RecPos: [%f %f %f]\n", Observer.X, Observer.Y, Observer.Z);
		mexPrintf("EarthTilt: %f Deg\n", EarthTilt);
		mexPrintf("EarthRadius: %f meters \n", EarthRadius);
		mexPrintf("Double Precision Computing\n");
#endif


		int RuntimeVersion = 0;
		int DriverVersion  = 0;
		int NumDevices     = 0;

		// ---------------------------------------------------
		// ------- GET CUDA RUNTIME AND DRIVER VERSION -------
		cudaRuntimeGetVersion(&RuntimeVersion);
		HANDLE_CUDA(cudaGetLastError());
		cudaDriverGetVersion(&DriverVersion);
		HANDLE_CUDA(cudaGetLastError());

		// ------------------------------------
		// ------- CHECK DRIVER SUPPORT -------
		if (RuntimeVersion > DriverVersion) {
			std::string Message = std::string("Cuda Runtime Version : ") + std::to_string(RuntimeVersion) + std::string("\n");
			Message += std::string("Cuda Driver Version  : ") + std::to_string(DriverVersion) + std::string("\n");
			Message += std::string("Internal Error. Insufficient Driver, please update Graphics Drivers to the latest version\n");
			mexErrMsgTxt(Message.c_str());
		}

		// -------------------------------------------
		// ------- CHECK FOR AMOUNT OF DEVICES -------
		cudaError_t Err = cudaGetDeviceCount(&NumDevices);
		if (Err == cudaErrorNoDevice) {
			std::string Message = std::string("Internal Error. No CUDA Capable Device Found!\n");
			mexErrMsgTxt(Message.c_str());
		}

		
#if DEBUG
		// --------------------------------------
		// ------- GET DEVICE PARAMETERS --------
		cudaDeviceProp DevProperties;
		HANDLE_CUDA(cudaGetDeviceProperties(&DevProperties, 0));
		mexPrintf("Found CUDA Capable GPU: %s \n", DevProperties.name);
#endif

		// -------------------------------------
		// ------- ALLOCATE DEVICE MEMORY ------
		double       *Dev_Rotation = nullptr;
		vec3<double> *Dev_EarthPos = nullptr;
		uint8_t      *Dev_Visible = nullptr;

		size_t RotationSize   = Nx * sizeof(double);
		size_t PositionSize   = Nx * sizeof(vec3<double>);
		size_t VisibilitySize = Nx * sizeof(uint8_t);

		size_t RequiredMem = RotationSize + PositionSize + VisibilitySize;
		size_t FreeMemory = 0;
		size_t TotalMemory = 0;
		HANDLE_CUDA(cudaMemGetInfo(&FreeMemory, &TotalMemory));
		if (RequiredMem > FreeMemory) {
			std::string Message = std::string(" Error Not enough graphic memory. Required:") + std::to_string(RequiredMem / (1024 * 1024)) + std::string(" MB | Available: ") + std::to_string(FreeMemory / (1024 * 1024)) + std::string(" MB\n");
			mexErrMsgTxt(Message.c_str());
		}


		HANDLE_CUDA(cudaMalloc((void**)&Dev_Rotation, RotationSize));
		HANDLE_CUDA(cudaMalloc((void**)&Dev_EarthPos, PositionSize));
		HANDLE_CUDA(cudaMalloc((void**)&Dev_Visible, VisibilitySize));
		HANDLE_CUDA(cudaMemset(Dev_Rotation, 0, RotationSize));
		HANDLE_CUDA(cudaMemset(Dev_EarthPos, 0, PositionSize));
		HANDLE_CUDA(cudaMemset(Dev_Visible, 0, VisibilitySize));

		// ----------------------------------
		// ------ ALLOCATE HOST MEMORY ------
		HANDLE_CUDA(cudaHostAlloc((void**)&Host_EarthPos, PositionSize, cudaHostAllocDefault));
		HANDLE_CUDA(cudaHostAlloc((void**)&Host_Rotation, RotationSize, cudaHostAllocDefault));

		// -----------------------------------------------
		// ------- SPECIFY KERNEL LAUNCHING PARAMS -------
		dim3 THREADS = dim3(1024, 1, 1);
		dim3 BLOCKS = dim3(Nx / 1024 + 1, 1, 1);

		// -------------------------------------------
		// ------ TRANSFORM TO ROW-MAJOR ORDER -------
		for (int i = 0; i < Nx; i++) {
			Host_Rotation[i] = In_Rot[i];
			Host_EarthPos[i].X = In_EarthPos[3 * i + 0];
			Host_EarthPos[i].Y = In_EarthPos[3 * i + 1];
			Host_EarthPos[i].Z = In_EarthPos[3 * i + 2];
		}

		// ----------------------------
		// ------ COPY TO DEVICE ------
		HANDLE_CUDA(cudaMemcpy(Dev_EarthPos, Host_EarthPos, PositionSize, cudaMemcpyHostToDevice));
		HANDLE_CUDA(cudaMemcpy(Dev_Rotation, Host_Rotation, RotationSize, cudaMemcpyHostToDevice));

		// ---------------------------------------
		// -------- Compute the Simulation -------
		Compute<double> << <BLOCKS, THREADS >> > (Dev_EarthPos, Sun, Observer, Dev_Rotation, EarthTilt, EarthRadius, Nx, Dev_Visible);
		HANDLE_CUDA(cudaGetLastError());

		// ---------------------------------------------
		// ------ ASSOCIATE OUTPUT AS A 2D Matrix ------
		const size_t OutputSize[] = { static_cast<size_t>(1) ,static_cast<size_t>(Nx) };
		plhs[0] = mxCreateNumericArray(2, OutputSize, mxUINT8_CLASS, mxREAL);
		uint8_t * Matlab_Visibility = static_cast<uint8_t*>(mxGetData(plhs[0]));

		// ---------------------------------------
		// ------ COPY RESULTS BACK TO HOST ------
		HANDLE_CUDA(cudaMemcpy(Matlab_Visibility, Dev_Visible, VisibilitySize, cudaMemcpyDeviceToHost));

		// ------------------------------------
		// ------ CLEAN UP DEVICE MEMORY ------
		if (Dev_Rotation != nullptr) { HANDLE_CUDA(cudaFree(Dev_Rotation)); Dev_Rotation = nullptr; }
		if (Dev_EarthPos != nullptr) { HANDLE_CUDA(cudaFree(Dev_EarthPos)); Dev_EarthPos = nullptr; }
		if (Dev_Visible != nullptr) { HANDLE_CUDA(cudaFree(Dev_Visible)); Dev_Visible = nullptr; }

		// -----------------------------------
		// ------ CLEAN UP HOST MEMORY -------
		if (Host_EarthPos != nullptr) { HANDLE_CUDA(cudaFreeHost(Host_EarthPos)); Host_EarthPos = nullptr; }
		if (Host_Rotation != nullptr) { HANDLE_CUDA(cudaFreeHost(Host_Rotation)); Host_Rotation = nullptr; }

	// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
	}else if((mxGetClassID(M_EarthPos) == mxSINGLE_CLASS) && (mxGetClassID(M_Rotation) == mxSINGLE_CLASS) && (mxGetClassID(M_SunPos) == mxSINGLE_CLASS) && (mxGetClassID(M_Observer) == mxSINGLE_CLASS)) {
	// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
		
		// -------------------------------------------
		// ------- Get Pointers to Data Arrays -------
		// NOTE: It is recommended to switch to mxGetSingles() in newer generation Matlabs
		float *In_EarthPos = static_cast<float*>(mxGetData(M_EarthPos));
		float *In_Rot      = static_cast<float*>(mxGetData(M_Rotation));
		float *In_Sun      = static_cast<float*>(mxGetData(M_SunPos));
		float *In_Obs      = static_cast<float*>(mxGetData(M_Observer));
		float *In_Add      = static_cast<float*>(mxGetData(M_Args));

		vec3<float> Sun;
		vec3<float> Observer;
		vec3<float>  *Host_EarthPos = nullptr;
		float *Host_Rotation = nullptr;
		float EarthTilt   = 23.48f;  // Degrees 
		float EarthRadius = 6378e3f; // Meters

		Sun.X = In_Sun[0];
		Sun.Y = In_Sun[1];
		Sun.Z = In_Sun[2];
		Observer.X = In_Obs[0];
		Observer.Y = In_Obs[1];
		Observer.Z = In_Obs[2];
		EarthTilt = In_Add[0];
		EarthRadius = In_Add[1];

		const mwSize* Dims = mxGetDimensions(M_EarthPos);
		int Nx = static_cast<int>(Dims[1]);
		int Ny = static_cast<int>(Dims[0]);

#if DEBUG
		
		mexPrintf("Nx: %d\n", Nx);
		mexPrintf("Ny: %d\n", Ny);
		mexPrintf("SunPos: [%f %f %f]\n", Sun.X, Sun.Y, Sun.Z);
		mexPrintf("RecPos: [%f %f %f]\n", Observer.X, Observer.Y, Observer.Z);
		mexPrintf("EarthTilt: %f Deg\n", EarthTilt);
		mexPrintf("EarthRadius: %f meters \n", EarthRadius);
		mexPrintf("Single Precision Computing\n");
#endif


		int RuntimeVersion = 0;
		int DriverVersion = 0;
		int NumDevices = 0;

		// ---------------------------------------------------
		// ------- GET CUDA RUNTIME AND DRIVER VERSION -------
		cudaRuntimeGetVersion(&RuntimeVersion);
		HANDLE_CUDA(cudaGetLastError());
		cudaDriverGetVersion(&DriverVersion);
		HANDLE_CUDA(cudaGetLastError());

		// -----------------------------------
		// ------ CHECK DRIVER SUPPORT -------
		if (RuntimeVersion > DriverVersion) {
			std::string Message = std::string("Cuda Runtime Version : ") + std::to_string(RuntimeVersion) + std::string("\n");
			Message += std::string("Cuda Driver Version  : ") + std::to_string(DriverVersion) + std::string("\n");
			Message += std::string("Internal Error. Insufficient Driver, please update Graphics Drivers to the latest version\n");
			mexErrMsgTxt(Message.c_str());
		}


		// -------------------------------------------
		// ------- CHECK FOR AMOUNT OF DEVICES -------
		cudaError_t Err = cudaGetDeviceCount(&NumDevices);
		if (Err == cudaErrorNoDevice) {
			std::string Message = std::string("Internal Error. No CUDA Capable Device Found!\n");
			mexErrMsgTxt(Message.c_str());
		}

		
#if DEBUG
		// --------------------------------------
		// ------- GET DEVICE PARAMETERS --------
		cudaDeviceProp DevProperties;
		HANDLE_CUDA(cudaGetDeviceProperties(&DevProperties, 0));
		mexPrintf("Found CUDA Capable GPU: %s \n", DevProperties.name);
#endif

		// -------------------------------------
		// ------- ALLOCATE DEVICE MEMORY ------
		float         *Dev_Rotation = nullptr;
		vec3<float>   *Dev_EarthPos = nullptr;
		uint8_t       *Dev_Visible  = nullptr;


		size_t RotationSize		= Nx * sizeof(float);
		size_t PositionSize		= Nx * sizeof(vec3<float>);
		size_t VisibilitySize	= Nx * sizeof(uint8_t);

		size_t RequiredMem = RotationSize + PositionSize + VisibilitySize;
		size_t FreeMemory = 0;
		size_t TotalMemory = 0;
		HANDLE_CUDA(cudaMemGetInfo(&FreeMemory, &TotalMemory));
		if (RequiredMem > FreeMemory) {
			std::string Message = std::string(" Error Not enough graphic memory. Required:") + std::to_string(RequiredMem / (1024 * 1024)) + std::string(" MB | Available: ") + std::to_string(FreeMemory / (1024 * 1024)) + std::string(" MB\n");
			mexErrMsgTxt(Message.c_str());
		}

		HANDLE_CUDA(cudaMalloc((void**)&Dev_Rotation, RotationSize));
		HANDLE_CUDA(cudaMalloc((void**)&Dev_EarthPos, PositionSize));
		HANDLE_CUDA(cudaMalloc((void**)&Dev_Visible, VisibilitySize));
		HANDLE_CUDA(cudaMemset(Dev_Rotation, 0, RotationSize));
		HANDLE_CUDA(cudaMemset(Dev_EarthPos, 0, PositionSize));
		HANDLE_CUDA(cudaMemset(Dev_Visible, 0, VisibilitySize));

		// ----------------------------------
		// ------ ALLOCATE HOST MEMORY ------
		HANDLE_CUDA(cudaHostAlloc((void**)&Host_EarthPos, PositionSize, cudaHostAllocDefault));
		HANDLE_CUDA(cudaHostAlloc((void**)&Host_Rotation, RotationSize, cudaHostAllocDefault));

		// ------------------------------------------------
		// -------- SPECIFY KERNEL LAUNCHING PARAMS -------
		dim3 THREADS = dim3(1024, 1, 1);
		dim3 BLOCKS = dim3(Nx / 1024 + 1, 1, 1);

		// -------------------------------------------
		// ------ TRANSFORM TO ROW-MAJOR ORDER -------
		for (int i = 0; i < Nx; i++) {
			Host_Rotation[i] = In_Rot[i];
			Host_EarthPos[i].X = In_EarthPos[3 * i + 0];
			Host_EarthPos[i].Y = In_EarthPos[3 * i + 1];
			Host_EarthPos[i].Z = In_EarthPos[3 * i + 2];
		}

		// ----------------------------
		// ------ COPY TO DEVICE ------
		HANDLE_CUDA(cudaMemcpy(Dev_EarthPos, Host_EarthPos, PositionSize, cudaMemcpyHostToDevice));
		HANDLE_CUDA(cudaMemcpy(Dev_Rotation, Host_Rotation, RotationSize, cudaMemcpyHostToDevice));

		// ---------------------------------------
		// -------- Compute The Simulation -------
		Compute<float> << <BLOCKS, THREADS >> > (Dev_EarthPos, Sun, Observer, Dev_Rotation, EarthTilt, EarthRadius, Nx, Dev_Visible);
		HANDLE_CUDA(cudaGetLastError());

		// ---------------------------------------------
		// ------ ASSOCIATE OUTPUT AS A 2D Matrix ------
		const size_t OutputSize[] = { static_cast<size_t>(1) ,static_cast<size_t>(Nx) };
		plhs[0] = mxCreateNumericArray(2, OutputSize, mxUINT8_CLASS, mxREAL);
		uint8_t * Matlab_Visibility = static_cast<uint8_t*>(mxGetData(plhs[0]));

		// ---------------------------------------
		// ------ COPY RESULTS BACK TO HOST ------
		HANDLE_CUDA(cudaMemcpy(Matlab_Visibility, Dev_Visible, VisibilitySize, cudaMemcpyDeviceToHost));

		// ------------------------------------
		// ------ CLEAN UP DEVICE MEMORY ------
		if (Dev_Rotation != nullptr) { HANDLE_CUDA(cudaFree(Dev_Rotation)); Dev_Rotation = nullptr; }
		if (Dev_EarthPos != nullptr) { HANDLE_CUDA(cudaFree(Dev_EarthPos)); Dev_EarthPos = nullptr; }
		if (Dev_Visible != nullptr) { HANDLE_CUDA(cudaFree(Dev_Visible)); Dev_Visible = nullptr; }

		// -----------------------------------
		// ------ CLEAN UP HOST MEMORY -------
		if (Host_EarthPos != nullptr) { HANDLE_CUDA(cudaFreeHost(Host_EarthPos)); Host_EarthPos = nullptr; }
		if (Host_Rotation != nullptr) { HANDLE_CUDA(cudaFreeHost(Host_Rotation)); Host_Rotation = nullptr; }

	}else {
		mexErrMsgTxt("Error Inputs must be ALL either Single or Double Precision!\n");
	}


	return;
}

